#include <stdint.h>
#include <assert.h>
#ifdef HAMMER_LLVM_BACKEND
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"
#include <llvm-c/Core.h>
#pragma GCC diagnostic pop
#include "../backends/llvm/llvm.h"
#endif
#include "parser_internal.h"

static HParseResult* parse_ch(void* env, HParseState *state) {
  uint8_t c = (uint8_t)(uintptr_t)(env);
  uint8_t r = (uint8_t)h_read_bits(&state->input_stream, 8, false);
  if (c == r) {
    HParsedToken *tok = a_new(HParsedToken, 1);    
    tok->token_type = TT_UINT; tok->uint = r;
    return make_result(state->arena, tok);
  } else {
    return NULL;
  }
}

static void desugar_ch(HAllocator *mm__, HCFStack *stk__, void *env) {
  HCFS_ADD_CHAR( (uint8_t)(uintptr_t)(env) );
}

static bool h_svm_action_ch(HArena *arena, HSVMContext *ctx, void* env) {
  // BUG: relies un undefined behaviour: int64_t is a signed uint64_t; not necessarily true on 32-bit
  HParsedToken *top = ctx->stack[ctx->stack_count-1];
  assert(top->token_type == TT_BYTES);
  uint64_t res = 0;
  for (size_t i = 0; i < top->bytes.len; i++)
    res = (res << 8) | top->bytes.token[i];   // TODO: Handle other endiannesses.
  top->uint = res; // possibly cast to signed through union
  top->token_type = TT_UINT;
  return true;
}

static bool ch_ctrvm(HRVMProg *prog, void* env) {
  uint8_t c = (uint8_t)(uintptr_t)(env);
  // TODO: Does this capture anything?
  h_rvm_insert_insn(prog, RVM_PUSH, 0);
  h_rvm_insert_insn(prog, RVM_MATCH, c | c << 8);
  h_rvm_insert_insn(prog, RVM_STEP, 0);
  h_rvm_insert_insn(prog, RVM_CAPTURE, 0);
  h_rvm_insert_insn(prog, RVM_ACTION, h_rvm_create_action(prog, h_svm_action_ch, env));
  return true;
}

#ifdef HAMMER_LLVM_BACKEND

static bool ch_llvm(HLLVMParserCompileContext *ctxt,
                    const HParser *p,
                    LLVMValueRef *res) {
  /* Build a new LLVM function to parse a character */

  /* Allocator */
  HAllocator *mm__ = NULL;
  /* Names */
  H_LLVM_DECLARE_NAME(ch_entry);
  H_LLVM_DECLARE_NAME(ch_success);
  H_LLVM_DECLARE_NAME(ch_end);
  H_LLVM_DECLARE_NAME(ch_read_bits);
  H_LLVM_DECLARE_NAME(ch_r);
  H_LLVM_DECLARE_NAME(ch_icmp);
  H_LLVM_DECLARE_NAME(ch_rv);
  /* Basic blocks we will be using */
  LLVMBasicBlockRef entry, success, end;
  /* Args for call to h_read_bits() */
  LLVMValueRef bits_args[3];
  /* Result of call to h_read_bits() */
  LLVMValueRef bits;
  /* Truncated result of call to h_read_bits() */
  LLVMValueRef r;
  /* Character to check for */
  uint8_t cval;
  /* Character to check for as an LLVM value */
  LLVMValueRef c;
  /* Result of LLVM comparison */
  LLVMValueRef icmp;
  /* Completed token */
  LLVMValueRef mr;
  /* Return value (a phi) */
  LLVMValueRef rv;
  /* Inputs to the phi */
  LLVMBasicBlockRef rv_phi_incoming_blocks[2];
  LLVMValueRef rv_phi_incoming_values[2];

  if (!res) return false;
  if (!p) return false;
  if (!ctxt) return false;

  /* Unpack the allocator */
  mm__ = ctxt->mm__;

  /*
   * Set up basic blocks: entry, success and failure branches, then exit
   */
  H_LLVM_COMPUTE_NAME(ch_entry, p);
  entry = LLVMAppendBasicBlock(ctxt->func, ch_entry_name);
  H_LLVM_COMPUTE_NAME(ch_success, p);
  success = LLVMAppendBasicBlock(ctxt->func, ch_success_name);
  H_LLVM_COMPUTE_NAME(ch_end, p);
  end = LLVMAppendBasicBlock(ctxt->func, ch_end_name);

  /* Basic block: entry */
  LLVMBuildBr(ctxt->builder, entry);
  LLVMPositionBuilderAtEnd(ctxt->builder, entry);

  /* Set up params for calls to h_read_bits() */
  bits_args[0] = ctxt->stream;
  bits_args[1] = LLVMConstInt(LLVMInt32Type(), 8, 0);
  bits_args[2] = LLVMConstInt(LLVMInt8Type(), 0, 0);

  /*
   * Call to h_read_bits()
   *
   * %read_bits = call i64 @h_read_bits(%struct.HInputStream_* %8,
   *                                    i32 8, i8 signext 0)
   */
  H_LLVM_COMPUTE_NAME(ch_read_bits, p);
  bits = LLVMBuildCall(ctxt->builder,
                       LLVMGetNamedFunction(ctxt->mod, "h_read_bits"),
                       bits_args, 3, ch_read_bits_name);
  /* %2 = trunc i64 %read_bits to i8 */
  H_LLVM_COMPUTE_NAME(ch_r, p);
  r = LLVMBuildTrunc(ctxt->builder, bits, LLVMInt8Type(), ch_r_name);

  /*
   * Check if h_read_bits succeeded
   *
   * %"c == r" = icmp eq i8 -94, %2 ; the -94 comes from c_
   */
  H_LLVM_COMPUTE_NAME(ch_icmp, p);
  cval = (uint8_t)(uintptr_t)(p->env);
  c = LLVMConstInt(LLVMInt8Type(), cval, 0);
  icmp = LLVMBuildICmp(ctxt->builder, LLVMIntEQ, c, r, ch_icmp_name);

  /*
   * Branch so success or failure basic block, as appropriate
   * br i1 %"c == r", label %ch_success, label %ch_fail
   */
  LLVMBuildCondBr(ctxt->builder, icmp, success, end);

  /* Basic block: success */
  LLVMPositionBuilderAtEnd(ctxt->builder, success);

  /* Make a token */
  h_llvm_make_tt_suint(ctxt, 8, 0, r, &mr);

  /* br label %ch_end */
  LLVMBuildBr(ctxt->builder, end);
  
  /* Basic block: end */
  LLVMPositionBuilderAtEnd(ctxt->builder, end);
  /*
   * %rv = phi %struct.HParseResult_.3* [ %make_result, %ch_success ],
   *       [ null, %ch_entry ]
   */
  H_LLVM_COMPUTE_NAME(ch_rv, p);
  rv = LLVMBuildPhi(ctxt->builder, ctxt->llvm_parseresultptr, ch_rv_name);
  rv_phi_incoming_blocks[0] = success;
  rv_phi_incoming_values[0] = mr;
  rv_phi_incoming_blocks[1] = entry;
  rv_phi_incoming_values[1] = LLVMConstNull(ctxt->llvm_parseresultptr);
  LLVMAddIncoming(rv, rv_phi_incoming_values, rv_phi_incoming_blocks, 2);
  /* Pass rv out to caller */
  *res = rv;

  /* Free any names we constructed */
  H_LLVM_FREE_NAME(ch_entry);
  H_LLVM_FREE_NAME(ch_success);
  H_LLVM_FREE_NAME(ch_end);
  H_LLVM_FREE_NAME(ch_read_bits);
  H_LLVM_FREE_NAME(ch_r);
  H_LLVM_FREE_NAME(ch_icmp);
  H_LLVM_FREE_NAME(ch_rv);

  return true;
}

#endif /* defined(HAMMER_LLVM_BACKEND) */

static const HParserVtable ch_vt = {
  .parse = parse_ch,
  .isValidRegular = h_true,
  .isValidCF = h_true,
  .desugar = desugar_ch,
  .compile_to_rvm = ch_ctrvm,
#ifdef HAMMER_LLVM_BACKEND
  .llvm = ch_llvm,
#endif
  .higher = false,
};

HParser* h_ch(const uint8_t c) {
  return h_ch__m(&system_allocator, c);
}
HParser* h_ch__m(HAllocator* mm__, const uint8_t c) {  
  return h_new_parser(mm__, &ch_vt, (void *)(uintptr_t)c);
}
