#ifdef HAMMER_LLVM_BACKEND
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wpedantic"
#include <llvm-c/Core.h>
#pragma GCC diagnostic pop
#include "../backends/llvm/llvm.h"
#endif
#include "parser_internal.h"

static HParseResult* parse_end(void *env, HParseState *state) {
  if (state->input_stream.index == state->input_stream.length) {
    HParseResult *ret = a_new(HParseResult, 1);
    ret->ast = NULL;
    return ret;
  } else {
    return NULL;
  }
}

static void desugar_end(HAllocator *mm__, HCFStack *stk__, void *env) {
  HCFS_ADD_END();
}

static bool end_ctrvm(HRVMProg *prog, void *env) {
  h_rvm_insert_insn(prog, RVM_EOF, 0);
  return true;
}

#ifdef HAMMER_LLVM_BACKEND

static bool end_llvm(HLLVMParserCompileContext *ctxt, 
                     const HParser *p,
                     LLVMValueRef *res) {
  /* Allocator */
  HAllocator *mm__ = NULL;
  /* Names */
  H_LLVM_DECLARE_NAME(end_entry);
  H_LLVM_DECLARE_NAME(end_success);
  H_LLVM_DECLARE_NAME(end_end);
  H_LLVM_DECLARE_NAME(end_index);
  H_LLVM_DECLARE_NAME(end_index_ptr);
  H_LLVM_DECLARE_NAME(end_length);
  H_LLVM_DECLARE_NAME(end_length_ptr);
  H_LLVM_DECLARE_NAME(end_icmp);
  H_LLVM_DECLARE_NAME(end_result_ptr);
  H_LLVM_DECLARE_NAME(end_rv);
  /* Basic blocks for parsing end */
  LLVMBasicBlockRef entry, success, end;
  /* Indices for GEP */
  LLVMValueRef gep_indices[2];
  /* Args to make_result() */
  LLVMValueRef make_result_args[2];
  /* Return from make_result() */
  LLVMValueRef result_ptr;
  /* Result of this parser */
  LLVMValueRef rv;
  /* Incoming basic blocks for phi */
  LLVMBasicBlockRef rv_phi_incoming_blocks[2];
  /* Incoming values for phi */
  LLVMValueRef rv_phi_incoming_values[2];

  if (!res) return false;
  if (!p) return false;
  if (!ctxt) return false;

  /* Unpack allocator */
  mm__ = ctxt->mm__;

  /* Set up some basic blocks */
  H_LLVM_COMPUTE_NAME(end_entry, p);
  entry = LLVMAppendBasicBlock(ctxt->func, end_entry_name);
  H_LLVM_COMPUTE_NAME(end_success, p);
  success = LLVMAppendBasicBlock(ctxt->func, end_success_name);
  H_LLVM_COMPUTE_NAME(end_end, p);
  end = LLVMAppendBasicBlock(ctxt->func, end_end_name);

  /* Basic block: entry */
  LLVMBuildBr(ctxt->builder, entry);
  LLVMPositionBuilderAtEnd(ctxt->builder, entry);

  /*
   * This needs to test if we're at the end of the input stream by
   * comparing the index and length fields; build a struct GEP to
   * get at their values.
   */

  /* The struct itself */
  gep_indices[0] = LLVMConstInt(LLVMInt32Type(), 0, 0);
  /* The index field (see HInputStream in internal.h */
  gep_indices[1] = LLVMConstInt(LLVMInt32Type(), 2, 0);
  /* GEP */
  H_LLVM_COMPUTE_NAME(end_index_ptr, p);
  LLVMValueRef index_ptr = LLVMBuildGEP(ctxt->builder,
      ctxt->stream, gep_indices, 2, end_index_ptr_name);

  /* The length field (see HInputStream in internal.h */
  gep_indices[1] = LLVMConstInt(LLVMInt32Type(), 3, 0);
  /* GEP */
  H_LLVM_COMPUTE_NAME(end_length_ptr, p);
  LLVMValueRef length_ptr = LLVMBuildGEP(ctxt->builder,
      ctxt->stream, gep_indices, 2, end_length_ptr_name);

  /* Now load them */
  H_LLVM_COMPUTE_NAME(end_index, p);
  LLVMValueRef index = LLVMBuildLoad(ctxt->builder,
      index_ptr, end_index_name);
  H_LLVM_COMPUTE_NAME(end_length, p);
  LLVMValueRef length = LLVMBuildLoad(ctxt->builder,
      length_ptr, end_length_name);

  /* Compare */
  H_LLVM_COMPUTE_NAME(end_icmp, p);
  LLVMValueRef icmp =
    LLVMBuildICmp(ctxt->builder, LLVMIntEQ, index, length, end_icmp_name);
  /* Branch on comparison */
  LLVMBuildCondBr(ctxt->builder, icmp, success, end);

  /* Basic block: success */
  LLVMPositionBuilderAtEnd(ctxt->builder, success);
  /* Set up a call to h_arena_malloc() to get an HParseResult */
  H_LLVM_COMPUTE_NAME(end_result_ptr, p);
  make_result_args[0] = ctxt->arena;
  make_result_args[1] = LLVMConstNull(ctxt->llvm_parsedtokenptr);
  result_ptr = LLVMBuildCall(ctxt->builder,
      LLVMGetNamedFunction(ctxt->mod, "make_result"),
      make_result_args, 2, end_result_ptr_name);

  /* Branch to end */
  LLVMBuildBr(ctxt->builder, end);

  /* Basic block: end */
  LLVMPositionBuilderAtEnd(ctxt->builder, end);
  /* Set up a phi depending on whether we have a token or not */
  H_LLVM_COMPUTE_NAME(end_rv, p);
  rv = LLVMBuildPhi(ctxt->builder, ctxt->llvm_parseresultptr, end_rv_name);
  rv_phi_incoming_blocks[0] = success;
  rv_phi_incoming_values[0] = result_ptr;
  rv_phi_incoming_blocks[1] = entry;
  rv_phi_incoming_values[1] = LLVMConstNull(ctxt->llvm_parseresultptr);
  LLVMAddIncoming(rv, rv_phi_incoming_values, rv_phi_incoming_blocks, 2);

  /* Pass it back out to the caller */
  *res = rv;

  /* Free names */
  H_LLVM_FREE_NAME(end_entry);
  H_LLVM_FREE_NAME(end_success);
  H_LLVM_FREE_NAME(end_end);
  H_LLVM_FREE_NAME(end_index);
  H_LLVM_FREE_NAME(end_index_ptr);
  H_LLVM_FREE_NAME(end_length);
  H_LLVM_FREE_NAME(end_length_ptr);
  H_LLVM_FREE_NAME(end_icmp);
  H_LLVM_FREE_NAME(end_result_ptr);
  H_LLVM_FREE_NAME(end_rv);

  return true;
}

#endif /* defined(HAMMER_LLVM_BACKEND) */

static const HParserVtable end_vt = {
  .parse = parse_end,
  .isValidRegular = h_true,
  .isValidCF = h_true,
  .desugar = desugar_end,
  .compile_to_rvm = end_ctrvm,
#ifdef HAMMER_LLVM_BACKEND
  .llvm = end_llvm,
#endif
  .higher = false,
};

HParser* h_end_p() {
  return h_end_p__m(&system_allocator);
}

HParser* h_end_p__m(HAllocator* mm__) { 
  return h_new_parser(mm__, &end_vt, NULL);
}
